-- written by DaddelZeit
-- DO NOT USE WITHOUT PERMISSION

local M = {}

local imguiUtils = require("ui/imguiUtils")
local icons = {}
local windowOpen = false
local im = ui_imgui
local allowMultiSelec = false
local style = im.GetStyle()

local currentPath = "/art/zeitRadio"
local filetypeString = "*"
local fileCache = nil
local queueRefresh = false
local callbackFunction = nop
local options = {file_exist = true, select_folder = false, suffix="*",pattern=nil}
local selectedFile = nil
local selectedFiles = {}
local lastSelectedFile = nil
local searchText = im.ArrayChar(512, "")

local userlang = settings.getValue("userLanguage")
local translationFile = "/art/zeitRadioGui/explorer_"..userlang..".json"
log("D", "uiLanguage", "Attempting to load file: "..translationFile)
local translations = jsonReadFile("/art/zeitRadioGui/explorer_"..userlang..".json") or jsonReadFile("/art/zeitRadioGui/explorer_en-US.json") or {}
local function translate(id)
    return translations[id] or id
end

local function toggleUI()
    windowOpen = not windowOpen
end

local function accept(data)
    local isok, msg = pcall(callbackFunction, data)
    if not isok then
        log("E", "", "Callback failed: "..dumps(msg))
    end

    toggleUI()
end

local function sendToAccept()
    if not selectedFile then return end
    if not allowMultiSelec then
        if selectedFile.path then
            accept(selectedFile.path)
        end
    else
        if next(selectedFiles) then
            accept(selectedFiles)
        end
    end
end

local function getFiles(localcachepath)
    currentPath = localcachepath
    fileCache = {}
    if currentPath:sub(1,1) ~= "/" then currentPath = "/" .. currentPath end
    if currentPath:sub(-1) ~= "/" then currentPath = currentPath.."/" end
    currentPath = currentPath:gsub("/+", "/")   -- Reducing multiple slashes to avoid trimmed list names
    local pathLen = string.len(currentPath) + 1
    -- get entries for current path
    local files = FS:findFiles(localcachepath, '*', 0, false, true) -- TODO : feature #4209
    local pattern = nil

    local validName
    for _, fn in ipairs(files) do
        local s = FS:stat(fn)
        s.name = string.sub(fn, pathLen)
        s.name = string.gsub(s.name, "/(.*)", "%1") -- strip leading /
        validName = true

        if not options.select_folder and (type(pattern) == "table" or (type(pattern) == 'string' and pattern ~= "*")) then
            if s.filetype == "file" then
                validName = checkPattern(s, pattern)
            end
        end

        if s.filetype == "file" then
            s.path = fn
            _, s.nameWithoutExt, s.extension = path.splitWithoutExt(fn)
            s.extension = string.lower(s.extension)
            if s.extension == "jpg" or s.extension == "png" or s.extension == "jpeg" or s.extension == "gif" then
                s.texObj = imguiUtils.texObj(s.path)
            end
        end

        if validName then
        if s.filetype == 'dir' then s.filesize = nil end
            table.insert(fileCache, s)
        end
    end
    -- sort them
    table.sort(fileCache, function(a, b)
        if a.filetype ~= b.filetype then
        if a.filetype == nil then return false end
        if b.filetype == nil then return true end
        return a.filetype < b.filetype
        end
        return string.lower(a.name) < string.lower(b.name)
    end)
    return fileCache
end

local dblClick = false
local columns = {
    {
        name = "Filetype",
        visible = true,
        guiFn = function(file)
            local prevCursorPos = im.GetCursorPos()
            im.SetCursorPos(im.ImVec2(prevCursorPos.x+style.ItemSpacing.x/2, prevCursorPos.y+style.ItemSpacing.y/2))
            if file.filetype == "dir" then
                im.Image(icons["simgroup"].texId, im.ImVec2(16,16))
            elseif file.extension == "png" or file.extension == "jpg" or file.extension == "jpeg" or file.extension == "gif" then
                im.Image(icons["image"].texId, im.ImVec2(16,16))
            elseif file.extension == "mp3" or file.extension == "wav" or file.extension == "ogg" or file.extension == "flac" then
                im.Image(icons["music"].texId, im.ImVec2(16,16))
            else
                im.Image(icons["unknown"].texId, im.ImVec2(16,16))
            end
            im.SameLine()
            im.SetCursorPosY(prevCursorPos.y)
        end
    },
    {
        name = "Filename",
        visible = true,
        lockedVisibility = true,
        guiFn = function(file)
            -- File
            if file.filetype == "file" then
                im.SameLine()

                local fileFound = arrayFindValueIndex(selectedFiles, file)
                if im.Selectable1(file.name, fileFound or selectedFile == file) then
                    selectedFile = file

                    if allowMultiSelec then
                        local SHIFT = im.GetIO().KeyShift
                        local CTRL = im.GetIO().KeyCtrl
                        if not fileFound then
                            if CTRL then
                                table.insert(selectedFiles, file)
                                lastSelectedFile = selectedFile
                            elseif SHIFT then
                                local fromIndex = 0
                                local toIndex = 0
                                for k,v in ipairs(fileCache or {}) do
                                    if lastSelectedFile and v.path == lastSelectedFile.path then
                                        fromIndex = k
                                    end
                                    if v.path == file.path then
                                        toIndex = k
                                    end
                                end
                                local curIndex = fromIndex
                                if fromIndex > toIndex then curIndex = toIndex toIndex = fromIndex fromIndex = curIndex end
                                while curIndex <= toIndex do
                                    if fileCache and fileCache[curIndex].fileTypeAllowed then table.insert(selectedFiles, fileCache[curIndex]) end
                                    curIndex = curIndex+1
                                end
                            else
                                selectedFiles = {}
                                table.insert(selectedFiles, file)
                                lastSelectedFile = selectedFile
                            end
                        else
                            if CTRL then
                                for k,v in ipairs(selectedFiles) do
                                    if v.path == file.path then
                                        table.remove(selectedFiles, k)
                                    end
                                end
                            else
                                selectedFiles = {} -- empty entire table
                                table.insert(selectedFiles, file)
                                lastSelectedFile = selectedFile
                            end
                        end
                    else
                        lastSelectedFile = selectedFile
                    end
                end
            else
                im.SameLine()
                im.Selectable1(file.name, false)
                if im.IsItemHovered() and dblClick then
                    currentPath = currentPath.."/"..file.name
                    queueRefresh = true
                end
            end
        end
    },
    {
        name = "Filesize",
        visible = true,
        guiFn = function(file)
            if file.filetype == "file" then
                if file.filesize then
                    im.SameLine()
                    im.TextColored(im.ImVec4(1.0, 1.0, 1.0, 0.6), bytes_to_string(file.filesize))
                    --im.TextUnformatted(bytes_to_string(file.filesize))
                end
            end
        end
    },
    {
        name = "Date created",
        visible = false,
        guiFn = function(file)
            im.SameLine()
            if file.filetype == "file" and file.createtime then
                im.TextUnformatted(os.date("%Y/%m/%d %I:%M %p", file.createtime))
            end
        end
    },
    {
        name = "Date modified",
        visible = false,
        guiFn = function(file)
            im.SameLine()
            if file.filetype == "file" and file.modtime then
                im.TextUnformatted(os.date("%Y/%m/%d %I:%M %p", file.createtime))
            end
        end
    }
}

local function checkType(allowedTypes, fileExtension)
    if allowedTypes == "*" then return true end -- always allow if *
    allowedTypes = allowedTypes:gsub("%s", "").."|"

    local _, sepAmount = allowedTypes:gsub("|", "")
    local ranTimes = 0
    local lastExtension = ""
    local extension = allowedTypes

    while sepAmount > ranTimes do
        extension = allowedTypes:match("[^|]+")
        lastExtension = extension
        allowedTypes = allowedTypes:gsub(lastExtension.."[|]", "")
        ranTimes = ranTimes + 1
        if fileExtension and fileExtension:lower():match(extension:lower()) then
            return true
        end
    end

    return false
end

local function fileChild(size, searchTerm)
    im.BeginChild1("ZeitRadioFileExplorer##fileChild", size)
    for k, file in pairs(fileCache) do
        if not file.name:lower():match(searchTerm) then goto next end
        if file.filetype ~= 'dir' then
            local fileDisabled = false
            if fileCache[k].fileTypeAllowed == nil then
                fileCache[k].fileTypeAllowed = checkType(filetypeString, file.extension)
            end

            if fileCache[k].fileTypeAllowed == false then
                fileDisabled = true
                im.BeginDisabled()
            end

            local fileHovered
            for k2, column in ipairs(columns) do
                if column.visible and column.visible == true then
                    if k2 == 1 then
                        im.Separator()
                    end
                    column.guiFn(file)
                    fileHovered = fileHovered or im.IsItemHovered()
                    im.NextColumn()
                end
            end
            if fileHovered and file.texObj then
                im.BeginTooltip()
                local imgSize = file.texObj.size
                local resize = 256/imgSize.x
                im.Image(file.texObj.texId, im.ImVec2(imgSize.x*resize, imgSize.y*resize))
                im.EndTooltip()
            end

            if fileDisabled == true then
               im.EndDisabled()
            end
        else
            for k2, column in ipairs(columns) do
                if column.visible and column.visible == true then
                  if k2 == 1 then
                    im.Separator()
                  end
                  column.guiFn(file)
                  im.NextColumn()
                end
            end
        end
        ::next::
    end
    im.EndChild()
end

local function onUpdate(dt)
    if not windowOpen then return end
    im.StyleColorsDark()
    im.PushStyleVar1(im.StyleVar_WindowRounding, 0)
    im.PushStyleVar1(im.StyleVar_ChildBorderSize, 0)

    im.SetNextWindowSizeConstraints(im.ImVec2(750, 0), im.ImVec2(750, 500))
    im.Begin("ZeitRadioFileExplorer", nil, im.WindowFlags_NoCollapse + im.WindowFlags_NoResize + im.WindowFlags_NoTitleBar + im.WindowFlags_AlwaysAutoResize)

    if not fileCache then
        queueRefresh = true
    end

    if currentPath == "/" then
        im.BeginDisabled()
        im.Button(translate("button.back"))
        im.EndDisabled()
    else
        if im.Button(translate("button.back")) then
            currentPath = (currentPath:gsub("%b//$", "") or "").."/"
            queueRefresh = true
        end
    end
    im.SameLine()
    if im.Button(translate("button.refresh")) then
        queueRefresh = true
    end

    if queueRefresh then
        selectedFiles = {}
        getFiles(currentPath)
        queueRefresh = false
    end

    im.SameLine()
    im.InputText(translate("search"), searchText)

    local windowWidth = im.GetWindowWidth()
    im.Separator()
    dblClick = im.IsMouseDoubleClicked(0)
    fileChild(im.ImVec2(windowWidth - style.WindowPadding.x*2, 250), ffi.string(searchText):lower())
    im.Separator()

    im.Dummy(im.ImVec2(0, 3))
    im.SameLine(windowWidth - im.CalcTextSize(translate("button.open")).x - im.CalcTextSize(translate("button.close")).x - style.ItemInnerSpacing.x*4 - style.ItemSpacing.x*2 - style.WindowPadding.x)
    if im.Button(translate("button.open")) then
        sendToAccept()
    end

    im.SameLine(windowWidth - im.CalcTextSize(translate("button.close")).x - style.ItemInnerSpacing.x - style.ItemSpacing.x - style.WindowPadding.x)
    if im.Button(translate("button.close")) then
        toggleUI()
    end
    im.End()

    im.PopStyleVar()
    im.PopStyleVar()

    Engine.imgui.enableBeamNGStyle()
end

M.onUpdate = onUpdate
M.open = function(newtypes, allowMultipleSelection, callback)
    searchText = im.ArrayChar(512, "")
    currentPath = "/art/zeitRadio"
    callbackFunction = callback or nop
    allowMultiSelec = allowMultipleSelection
    filetypeString = newtypes or "*"
    getFiles(currentPath)
    toggleUI()
end

M.onExtensionLoaded = function()
    for k,v in ipairs(FS:findFiles("/art/zeitRadioGui/", "*.png\t*.jpg", 0, true, false)) do
        icons[fileBase(v)] = imguiUtils.texObj(v)
    end
end

return M