-- written by DaddelZeit
-- DO NOT USE WITHOUT PERMISSION

local M = {}

local im = ui_imgui
local imguiUtils = require("ui/imguiUtils")
local icons = {}

local radioChangedSinceStart = false
local isOldFormat = false
local queueSelfUnload

local trackTable = {}
local playlistTable = {}
local searchText = im.ArrayChar(512, "")
local whiteColor = im.GetColorU322(im.ImVec4(1,1,1,1))

local currentTrackName
local currentTrack
local currentTrackObjs

local currentListId
local currentList

local userlang = settings.getValue("userLanguage")
local translationFile = "/art/zeitRadioGui/assistant_"..userlang..".json"
log("D", "uiLanguage", "Attempting to load file: "..translationFile)
local translations = jsonReadFile(translationFile) or jsonReadFile("/art/zeitRadioGui/assistant_en-US.json") or {}
local function translate(id)
    return translations[id] or id
end

local function textMonospace(text)
    local pushFont = im.PushFont3("robotomono_regular")
    im.Text(text or "")
    if pushFont then im.PopFont() end
end

local function customCheckbox(name, bool)
    local imStyle = im.GetStyle()
    local prevCursorPos = im.GetCursorPos()
    im.SetCursorPos(im.ImVec2(im.GetCursorPosX() + imStyle.ItemSpacing.x/2, im.GetCursorPosY() + imStyle.ItemSpacing.y))
    local screenCursorPos = im.GetCursorScreenPos()

    local textSize = im.CalcTextSize(name)
    local hovering = im.IsMouseHoveringRect(screenCursorPos, im.ImVec2(screenCursorPos.x + 14 + textSize.x + imStyle.ItemSpacing.x*2, screenCursorPos.y + 14))
    im.Image(
        bool and (hovering and icons["checkbox_on_hovered"].texId or icons["checkbox_on"].texId)
        or (hovering and icons["checkbox_off_hovered"].texId or icons["checkbox_off"].texId),
        im.ImVec2(14,14)
    )
    im.SameLine()
    im.SetCursorPosY(prevCursorPos.y)
    im.Text(name)

    return hovering and im.IsMouseClicked(0)
end

local function iconButton(name, icon)
    local imStyle = im.GetStyle()
    local textAlign = im.ImVec2(imStyle.ButtonTextAlign.x, imStyle.ButtonTextAlign.y)
    local innerSpacing = im.ImVec2(imStyle.ItemInnerSpacing.x, imStyle.ItemInnerSpacing.y)

    imStyle.ButtonTextAlign = im.ImVec2(1,0.5)

    local textSize = im.CalcTextSize(name)
    local cursorPos = im.GetCursorScreenPos()
    local topLeft = im.ImVec2(cursorPos.x+(innerSpacing.x/2), cursorPos.y+(innerSpacing.y/2))
    local btmRigh = im.ImVec2(topLeft.x+textSize.y-(innerSpacing.x/4), topLeft.y+textSize.y-(innerSpacing.y/4))

    local real = im.Button(name, im.ImVec2(textSize.x+textSize.y+(innerSpacing.x*3), textSize.y+innerSpacing.y))
    im.ImDrawList_AddImage(im.GetWindowDrawList(), icons[icon].texId, topLeft, btmRigh, im.ImVec2(0,0), im.ImVec2(1,1), whiteColor)

    imStyle.ButtonTextAlign = textAlign

    return real
end

local function getDisplayName(key)
    local name = key
    if trackTable[key] then
        if trackTable[key].artist ~= "" and trackTable[key].name ~= "" then
            name = trackTable[key].artist.." - "..trackTable[key].name
        elseif trackTable[key].name ~= "" then
            name = trackTable[key].name
        elseif trackTable[key].artist ~= "" then
            name = trackTable[key].artist
        end
    end
    return name
end

local function getTrackTableUINames(tbl)
    local keys = table.new(#tbl, 0)
    local keysidx = 0
    for k in pairs(tbl) do
        keysidx = keysidx + 1
        keys[keysidx] = {k, getDisplayName(k)}
    end
    return keys
end


local function updateTrackSelection(id)
    currentTrackName = id
    currentTrack = trackTable[currentTrackName]
    currentTrackObjs = {
        name = im.ArrayChar(512, currentTrack.name),
        artist = im.ArrayChar(512, currentTrack.artist),
        volCoef = im.FloatPtr(currentTrack.volCoef or 1),
        duration = {
            min = math.floor(currentTrack.duration/60),
            sec = currentTrack.duration%60
        }
    }
    zeit_trackTableEditorAPI.setCurrentTrack(currentTrackName)
end

local function updatePlaylistSelection(id)
    currentListId = id
    currentList = playlistTable[currentListId]
    zeit_trackTableEditorAPI.setCurrentPlaylist(currentListId)
end

local function selectNoneTrack()
    currentTrackName = ""
    currentTrack = {}
    currentTrackObjs = {
        name = im.ArrayChar(512, ""),
        artist = im.ArrayChar(512, ""),
        volCoef = im.FloatPtr(1),
        duration = {
            min = 0,
            sec = 0
        }
    }
end

local function selectNoneList()
    currentListId = nil
    currentList = {}
end


local function renderSongOptions()
    local trackDirty = false

    if im.InputText(translate("song_options.name"), currentTrackObjs.name) then
        trackDirty = true
    end
    if im.InputText(translate("song_options.artist"), currentTrackObjs.artist) then
        trackDirty = true
    end

    im.Dummy(im.ImVec2(0,im.GetStyle().WindowPadding.y*2))

    im.BeginChild1("##ZeitRadioSongWrapper", im.ImVec2(im.GetWindowSize().x-im.GetStyle().WindowPadding.x*2, 60), false, im.WindowFlags_HorizontalScrollbar)
    if iconButton(translate("song_options.change_music"), "simgroup") then
        zeit_trackTableEditorAPI.openExplorerSingle("music")
        trackDirty = true
    end
    im.SameLine()
    textMonospace(currentTrack.url)
    if iconButton(translate("song_options.change_cover"), "simgroup") then
        zeit_trackTableEditorAPI.openExplorerSingle("cover")
        trackDirty = true
    end
    im.SameLine()
    textMonospace(currentTrack.coverUrl)
    im.EndChild()

    if im.Button("-##mins") then
        currentTrackObjs.duration.min = currentTrackObjs.duration.min - 1
        trackDirty = true
    end
    im.SameLine()
    textMonospace(tostring(currentTrackObjs.duration.min))
    im.SameLine()
    if im.Button("+##mins") then
        currentTrackObjs.duration.min = currentTrackObjs.duration.min + 1
        trackDirty = true
    end
    im.SameLine()
    im.Text(translate("song_options.minutes"))

    if im.Button("-##secs") then
        currentTrackObjs.duration.sec = currentTrackObjs.duration.sec - 1
        trackDirty = true
    end
    im.SameLine()
    textMonospace(tostring(currentTrackObjs.duration.sec))
    im.SameLine()
    if im.Button("+##secs") then
        currentTrackObjs.duration.sec = currentTrackObjs.duration.sec + 1
        trackDirty = true
    end
    im.SameLine()
    im.Text(translate("song_options.seconds"))

    im.Dummy(im.ImVec2(0,2))

    if im.SliderFloat(translate("song_options.volumecoef"), currentTrackObjs.volCoef, 0.0, 5.0) then
        trackDirty = true
    end

    if trackDirty then
        radioChangedSinceStart = true
        zeit_trackTableEditorAPI.saveTrack(currentTrackName,
        ffi.string(currentTrackObjs.name),
        ffi.string(currentTrackObjs.artist),
        currentTrackObjs.volCoef[0],
        currentTrackObjs.duration.min*60 + currentTrackObjs.duration.sec)
    end
end

local function renderPlaylist()
    if not currentListId then return end
    local listDirty = false
    local sortedTrackTable = isOldFormat and {} or getTrackTableUINames(trackTable)
    table.sort(sortedTrackTable, function(a, b)
        return a[2] < b[2]
    end)

    local function entryRender(key)
        im.SameLine()
        im.InputText(translate("song.search"), searchText)
        im.Dummy(im.ImVec2(0,5))
        for _,track in pairs(sortedTrackTable) do
            if track[2]:lower():match(ffi.string(searchText):lower()) then
                if im.Selectable1(track[2], track[1] == v) then
                    currentList[key] = track[1]
                    listDirty = true
                end
            end
        end
    end

    for k,v in ipairs(currentList) do
        im.Text(k..". ")
        im.SameLine()
        if im.BeginCombo("##Playlist"..k, v) then
            entryRender(k)
            im.EndCombo()
        end
        im.SameLine()
        if im.Button(translate("playlist_options.remove").."##Playlist"..k) then
            table.remove(currentList, k)
            listDirty = true
        end
    end

    im.Dummy(im.ImVec2(0,5))
    if im.BeginCombo("##Playlist"..#currentList+1, translate("playlist_options.addsong")) then
        entryRender(#currentList+1)
        im.EndCombo()
    end

    if listDirty then
        radioChangedSinceStart = true
        zeit_trackTableEditorAPI.updatePlaylist(currentList)
    end
end


local renderList = {
    header = function()
        if iconButton(translate("button.open_folder"), "simgroup") then
            Engine.Platform.exploreFolder(zeit_trackTableEditorFileManager.filepath)
        end
        im.SameLine()
        if radioChangedSinceStart then
            if iconButton(translate("button.apply"), "checkbox_on") then
                core_vehicle_manager.reloadAllVehicles()
                radioChangedSinceStart = false
            end
            im.SameLine()
        end
        if im.Selectable1(translate("button.open_json"), false, 0, im.ImVec2(im.CalcTextSize(translate("button.open_json")).x, 18)) then
            Engine.Platform.exploreFolder("/settings/zeit/radio/")
        end
    end,
    song = function()
        if not currentListId then
            iconButton(translate("button.add_song"), "new_h")
            if im.IsItemHovered() then
                im.BeginTooltip()
                im.TextColored(im.ImVec4(1,0,0,1), translate("song.import_instruction"))
                im.EndTooltip()
            end
        else
            if iconButton(translate("button.add_song"), "new_h") then
                zeit_trackTableEditorAPI.openExplorer()
                radioChangedSinceStart = true
            end
            if im.IsItemHovered() then
                im.BeginTooltip()
                im.Text(translate("song.import_desc"))
                im.Text(translate("song.import_desc_2"))
                im.EndTooltip()
            end
        end
        im.SameLine()

        if currentTrackName == "" then
            iconButton(translate("button.delete_song"), "clear-btn_d")
            if im.IsItemHovered() then
                im.BeginTooltip()
                im.TextColored(im.ImVec4(1,0,0,1), translate("song.delete_instruction"))
                im.EndTooltip()
            end
        else
            if iconButton(translate("button.delete_song"), "clear-btn_d") then
                if currentTrackName then zeit_trackTableEditorAPI.deleteSong(currentTrackName) end
                selectNoneTrack()
                radioChangedSinceStart = true
            end
            if im.IsItemHovered() then
                im.BeginTooltip()
                im.Text(translate("song.delete_desc"))
                im.EndTooltip()
            end
        end

        im.SameLine()
        local sortedTrackTable = isOldFormat and {} or getTrackTableUINames(trackTable)
        table.sort(sortedTrackTable, function(a, b)
            return a[2] < b[2]
        end)
        if im.BeginCombo(translate("song").."##ZeitRadio", currentTrackName, im.ComboFlags_HeightLarge) then
            im.SameLine()
            im.InputText(translate("song.search"), searchText)
            im.Dummy(im.ImVec2(0,5))
            for _,track in ipairs(sortedTrackTable) do
                if track[2]:lower():match(ffi.string(searchText):lower()) then
                    if im.Selectable1(track[2], track[1] == currentTrackName) then
                        updateTrackSelection(track[1])
                    end
                end
            end
            im.EndCombo()
        end

        if currentTrackName == "" then im.BeginDisabled() end
        im.Dummy(im.ImVec2(0,2))
        renderSongOptions()
        im.Dummy(im.ImVec2(0,2))
        if currentTrackName == "" then im.EndDisabled() end
    end,
    playlist = function()
        if iconButton(translate("button.add_playlist"), "new_h") then
            currentListId = #playlistTable+1
            currentList = {}
            zeit_trackTableEditorAPI.setCurrentPlaylist(currentListId)
            zeit_trackTableEditorAPI.updatePlaylist({})
            radioChangedSinceStart = true
        end
        if im.IsItemHovered() then
            im.BeginTooltip()
            im.Text(translate("playlist.add_desc"))
            im.EndTooltip()
        end

        im.SameLine()
        if not currentListId then
            iconButton(translate("button.delete_playlist"), "clear-btn_d")
            if im.IsItemHovered() then
                im.BeginTooltip()
                im.TextColored(im.ImVec4(1,0,0,1), translate("playlist.delete_instruction"))
                im.EndTooltip()
            end
        else
            if iconButton(translate("button.delete_playlist"), "clear-btn_d") then
                if currentListId then
                    if currentListId == 1 then
                        zeit_trackTableEditorAPI.updatePlaylist({})
                    else
                        zeit_trackTableEditorAPI.deletePlaylist(currentListId)
                    end
                    radioChangedSinceStart = true
                end
                selectNoneList()
            end
            if im.IsItemHovered() then
                im.BeginTooltip()
                im.Text(translate("playlist.delete_desc"))
                im.EndTooltip()
            end
        end

        im.SameLine()
        if im.BeginCombo(translate("playlist").."##ZeitRadio", tostring(currentListId or ""), im.ComboFlags_HeightLarge) then
            for list, _ in pairs(playlistTable) do
                if im.Selectable1(tostring(list), list == currentListId) then
                    updatePlaylistSelection(list)
                end
            end
            im.EndCombo()
        end

        im.BeginChild1(tostring(currentListId or 0).."##ZeitRadioPlaylist", im.ImVec2(im.GetWindowSize().x-im.GetStyle().WindowPadding.x*2, 175), true)
        im.Dummy(im.ImVec2(0,2))
        renderPlaylist()
        im.Dummy(im.ImVec2(0,2))
        im.EndChild()
    end,
    settings = function()
        local imStyle = im.GetStyle()
        local prevCursorPos = im.GetCursorPos()
        im.SetCursorPos(im.ImVec2(im.GetCursorPosX()+imStyle.ItemSpacing.x/2, im.GetCursorPosY()+imStyle.ItemSpacing.y))
        im.Image(icons["settings"].texId, im.ImVec2(14,14))
        im.SameLine()
        im.SetCursorPosY(prevCursorPos.y)
        im.Text(translate("settings"))

        -- these are hardcoded...i mean theyre settings
        im.BeginChild1("Settings", im.ImVec2(im.GetWindowSize().x-im.GetStyle().WindowPadding.x*2, 85), true)
        local setting = settings.getValue("zeit_radio_thirdperson")
        local active = setting == nil and false or setting
        if customCheckbox(translate("setting.thirdperson"), active) then
            if active then
                settings.setValue("zeit_radio_thirdperson", false, true)
            else
                settings.setValue("zeit_radio_thirdperson", true, true)
            end
        end

        setting = settings.getValue("zeit_radio_rememberafterreset")
        active = setting == nil and true or setting
        if customCheckbox(translate("setting.remember"), active) then
            if active then
                settings.setValue("zeit_radio_rememberafterreset", false, true)
            else
                settings.setValue("zeit_radio_rememberafterreset", true, true)
            end
        end

        setting = settings.getValue("zeit_radio_loopplaylist")
        active = setting == nil and false or setting
        if customCheckbox(translate("setting.loop"), active) then
            if active then
                settings.setValue("zeit_radio_loopplaylist", false, true)
            else
                settings.setValue("zeit_radio_loopplaylist", true, true)
            end
        end

        setting = settings.getValue("zeit_radio_shuffleplaylist")
        active = setting == nil and false or setting
        if customCheckbox(translate("setting.shuffle"), active) then
            if active then
                settings.setValue("zeit_radio_shuffleplaylist", false, true)
            else
                settings.setValue("zeit_radio_shuffleplaylist", true, true)
            end
        end
        im.EndChild()
    end
}

local function onUpdate(dtReal)
    if queueSelfUnload then
        extensions.unload("zeit_tracktableEditor")
        queueSelfUnload = false
        return
    end

    im.StyleColorsDark()
    im.PushStyleVar1(im.StyleVar_WindowRounding, 0)
    im.PushStyleVar1(im.StyleVar_ChildBorderSize, 0)

    im.Begin("Zeit Radio Assistant", nil, im.WindowFlags_AlwaysAutoResize)

    if isOldFormat then
        im.Text(translate("upgrade.oldformat"))
        im.Text(translate("upgrade.instruction"))
        if im.Button(translate("button.update")) then
            trackTable, playlistTable = zeit_trackTableEditorAPI.updateFormat()
            isOldFormat = false
        else
            im.BeginDisabled()
        end
        im.Separator()
    end

    renderList.header()
    im.Separator()
    renderList.song()
    im.Separator()
    renderList.playlist()
    if isOldFormat then im.EndDisabled() end
    im.Separator()
    renderList.settings()

    im.PopStyleVar()
    im.PopStyleVar()
    im.End()
    Engine.imgui.enableBeamNGStyle()
end

local function onExtensionLoaded()
    if not FS:fileExists("/settings/zeit/radio/trackTable.json") then
        queueSelfUnload = true
    end
    selectNoneTrack()
    selectNoneList()

    extensions.load("zeit_trackTableEditorAPI")
    setExtensionUnloadMode(zeit_trackTableEditorAPI, "manual")

    trackTable, playlistTable, isOldFormat = zeit_trackTableEditorAPI.initAPI()

    for k,v in ipairs(FS:findFiles("/art/zeitRadioGui/", "*.png\t*.jpg", 0, true, false)) do
        icons[fileBase(v)] = imguiUtils.texObj(v)
    end
end

local function onExtensionUnloaded()
    extensions.unload("zeit_trackTableEditorAPI")
end

M.onUpdate = onUpdate
M.updateTrackSelection = updateTrackSelection
M.updatePlaylistSelection = updatePlaylistSelection
M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded

return M