-- written by DaddelZeit
-- DO NOT USE WITHOUT PERMISSION

local M = {}

local trackTable
local playlistTable

local filePathsToProcess = {}

local trackName
local currentTrack
local currentList = 2

local function _dialogCallback(files)
  for _,v in ipairs(files) do
    table.insert(filePathsToProcess, v.path)
  end
end

local function _dialogCallbackSingleImage(file)
  trackTable[currentTrack].coverUrl = file:startswith("/") and file:sub(2) or file

  M.saveTrack(currentTrack,
  trackTable[currentTrack].name,
  trackTable[currentTrack].artist,
  trackTable[currentTrack].volCoef,
  trackTable[currentTrack].duration)
end

local function _dialogCallbackSingleMusic(file)
  trackTable[currentTrack].url = file:startswith("/") and file:sub(2) or file

  M.saveTrack(currentTrack,
  trackTable[currentTrack].name,
  trackTable[currentTrack].artist,
  trackTable[currentTrack].volCoef,
  trackTable[currentTrack].duration)
end

--[[
local function getID3(filename)
  local ID3 = {}
  local file = assert(io.open(filename,"rb"))
  file:seek("end")
  file:seek("cur",-128)
  ID3.TAG = file:read(3):gsub("%z", "")
  ID3.TITLE = file:read(30):gsub("%z", "")
  ID3.INTERPRET = file:read(30):gsub("%z", "")
  ID3.ALBUM = file:read(30):gsub("%z", "")
  ID3.YEAR = file:read(4):gsub("%z", "")
  ID3.COMMENT = file:read(30):gsub("%z", "")
  file:close()
  for k in pairs(ID3) do
    if not ID3[k] then table.remove(ID3,k) end
  end
  return ID3
end
]]--

local function generate(file)
  if not file then return end

  file = file:match("art/.+")
  local extension = file:match("[^.]+$")

  local duration
  local tempSource = Engine.Audio.playOnce("AudioGui", file, {
    volume = 0,
    pitch = 0,
    fadeInTime = 0,
    fadeOutTime = 0,
  })

  if tempSource then
    duration = math.ceil(tempSource.len)
    scenetree.findObject(tempSource.sourceId):deleteObject() -- stop immediately after and also remove
  else
    -- import failed!
    log('E', "tracktableEditorAPI.generate", "File import failed: unable to create source. File format incorrect?")
  end

  trackName = file:gsub("art/zeitRadio/", ""):gsub("."..extension, "")
  local trackNameUI = trackName:gsub(".+/", "")
  trackName = trackName:gsub('[%p%c%s]', '_'):lower()

  -- dirty check if there actually are proper tags
  local id3Tags = zeit_id3 and zeit_id3.readtags(file) or {}

  local cover = "art/zeitRadio/empty.png"
  if next(id3Tags) then
    if id3Tags.cover then
      local fileType = id3Tags.cover:match("[^image/].+[^\x03]"):sub(1,4):gsub("[^%a+]", "")
      -- nobody calls jpg files jpeg anymore
      local path = file:gsub(extension, fileType == "jpeg" and "jpg" or "png")
      path = zeit_trackTableEditorFileManager.filepath:gsub("art/zeitRadio/", "")..path:gsub('[%s]', '_')

      if not FS:fileExists(path) then
        if fileType == "png" then
          local fileData = id3Tags.cover:sub(id3Tags.cover:find("PNG")-1, id3Tags.cover:find("IEND")+7)
          writeFile(path, fileData)
          cover = path
        else
          local fileData = id3Tags.cover:sub(id3Tags.cover:find("JFIF")-6, id3Tags.cover:len())
          writeFile(path, fileData)
          cover = path
        end
      end
    end

    for type,val in pairs(id3Tags) do
      id3Tags[type] = --[[val:gsub("%z", "")]] val:gsub("[^%a%d%p%s]", "")
    end
  end

  cover = cover:gsub('[%s]', '_')
  if not trackTable[trackName] then
    trackTable[trackName] = {
      name     = id3Tags.title or trackNameUI,
      artist   = id3Tags.artist or "",
      url      = file:startswith("/") and file:sub(1, file:len()) or file,
      coverUrl = cover:startswith("/") and cover:sub(1, cover:len()) or cover,
      duration = tonumber(duration) or 0,
      volCoef  = 2
    }
  elseif trackTable[trackName].duration == (0 or nil) then
    trackTable[trackName].duration = tonumber(duration) or 0
  end

  return trackName
end
M.generate = generate

local function onUpdate(dt)
  if not filePathsToProcess[1] then return end
  local newTrackName = generate(filePathsToProcess[1])
  table.remove(filePathsToProcess, 1)

  if #filePathsToProcess == 0 then
    zeit_tracktableEditor.updateTrackSelection(newTrackName)
    guihooks.trigger('app:waiting', false)
  end

  local insertList = playlistTable[currentList]
  if not arrayFindValueIndex(insertList, newTrackName) then
    table.insert(insertList, newTrackName)
    zeit_tracktableEditor.updatePlaylistSelection(currentList)
  end

  jsonWriteFile("settings/zeit/radio/trackTable.json", trackTable, true)
  jsonWriteFile("settings/zeit/radio/playlists.json", playlistTable, true)
end

local function cleanPlaylists()
  for listid,list in ipairs(playlistTable) do
    for i=#list, 1, -1 do
      if trackTable[list[i]] == nil then
        table.remove(playlistTable[listid], key)
      end
    end
  end
end

local function initAPI()
  trackTable = jsonReadFile("settings/zeit/radio/trackTable.json") or {}
  playlistTable = jsonReadFile("settings/zeit/radio/playlists.json") or {}

  extensions.load("zeit_trackTableEditorFileManager")
  setExtensionUnloadMode(zeit_trackTableEditorFileManager, "manual")
  extensions.load("zeit_radioFileDialog")
  setExtensionUnloadMode(zeit_radioFileDialog, "manual")
  extensions.load("zeit_id3")
  setExtensionUnloadMode(zeit_id3, "manual")

  local oldFormat = trackTable.playlists and not trackTable.playlists.name
  if not oldFormat then
    cleanPlaylists()
    jsonWriteFile("settings/zeit/radio/playlists.json", playlistTable, true)
  end
  return trackTable, playlistTable, oldFormat
end

local function onExtensionUnloaded()
  extensions.unload("zeit_trackTableEditorFileManager")
  extensions.unload("zeit_radioFileDialog")
  extensions.unload("zeit_id3")
end

local function setCurrentTrack(newTrack)
  currentTrack = newTrack
end

local function setCurrentPlaylist(newList)
  currentList = newList
end

local function updateFormat()
  local _trackTable = jsonReadFile("settings/zeit/radio/trackTable.json") or {}
  local _playlistTable = {}

  for k,v in pairs(_trackTable.playlists) do
    table.insert(_playlistTable, deepcopy(v))
  end
  _trackTable.playlists = nil

  trackTable = _trackTable
  playlistTable = _playlistTable

  cleanPlaylists()
  jsonWriteFile("settings/zeit/radio/trackTable.json", trackTable, true)
  jsonWriteFile("settings/zeit/radio/playlists.json", playlistTable, true)

  return trackTable, playlistTable
end

function M.deleteSong(track)
  for listId, list in ipairs(playlistTable) do
    local arrayIndex = arrayFindValueIndex(list, track)
    if arrayIndex then
      table.remove(playlistTable[listId], arrayIndex)
    end
  end

  trackTable[track] = nil
  jsonWriteFile("settings/zeit/radio/trackTable.json", trackTable, true)
  jsonWriteFile("settings/zeit/radio/playlists.json", playlistTable, true)
end

function M.deletePlaylist(id)
  table.remove(playlistTable, id)
  jsonWriteFile("settings/zeit/radio/playlists.json", playlistTable, true)
end

function M.updatePlaylist(data)
  local newPlaylist = {}
  for k,v in pairs(data) do
    newPlaylist[tonumber(k)] = v
  end

  -- convert to array
  local newPlaylistArray = {}
  for _, v in pairs(newPlaylist) do
    table.insert(newPlaylistArray, v)
  end

  playlistTable[currentList] = newPlaylistArray

  jsonWriteFile("settings/zeit/radio/playlists.json", playlistTable, true)
end

function M.getSongData(track)
  if track then
    return trackTable[track]
  else
    return trackTable[currentTrack]
  end
end

function M.saveTrack(key, name, artist, volCoef, duration)
  trackTable[key].name = name
  trackTable[key].artist = artist
  trackTable[key].volCoef = tonumber(volCoef)
  if duration == "NaN" or duration == "undefined" then
    trackTable[key].duration = 0
  else
    trackTable[key].duration = tonumber(duration) or 0
  end

  jsonWriteFile("settings/zeit/radio/trackTable.json", trackTable, true)
end

function M.dumpTrackData(bool)
  if bool then return trackTable else dump(trackTable) end
end

-- explorer open functions
M.openExplorer = function()
  zeit_radioFileDialog.open(
    "mp3 | wav | ogg",
    true,
    _dialogCallback
  )
end

M.openExplorerSingle = function(type)
  if type == "music" then
    zeit_radioFileDialog.open(
      "mp3 | wav | ogg",
      false,
      _dialogCallbackSingleMusic
    )
  elseif type == "cover" then
    zeit_radioFileDialog.open(
      "jpg | jpeg | png",
      false,
      _dialogCallbackSingleImage
    )
  end
end

M.initAPI = initAPI
M.updateFormat = updateFormat
M.setCurrentTrack = setCurrentTrack
M.setCurrentPlaylist = setCurrentPlaylist
M.onExtensionUnloaded = onExtensionUnloaded
M.onUpdate = onUpdate

return M