-- written by DaddelZeit
-- DO NOT USE WITHOUT PERMISSION

local M = {}

local dontCamInCarCheck = false
local doSerialize = true

local isInCar = false
local lastIsInCar = false
local camDistToCar = 0
local camPos
local vehPos
local driverCamId

local radioVolume = 5

local trackTable = {}
local playlists = {}
local amountOfSongsInTable = 0
local amountOfPlaylists = 0

local songUnloadQueue = {} -- done to reduce lag
local currentPlaylistIndex = 1 -- 1 is 1st playlist
local currentPlaylistIds = {}
local currentlyPlayingSongIndex = 1 -- 1 is off: "_tempData"
local currentlyPlayingSongData = {}

local soundNode = nil
local songTimer = 0
local prePauseRadioVolume = 0
local isPaused = false
local pitch = 1 -- used for pause; goofy solution but works, surprisingly

local updateUITimer = 0

local hideMessages = false

local function tuneRadioVolume(val, showMessages, fullVal, setByPause) -- fullVal is a bool!
    if not fullVal then
        radioVolume = math.min(math.max(radioVolume + val, 0), 20)
    else
        radioVolume = math.min(math.max(val, 0), 20)
    end
    if not currentlyPlayingSongData or not currentlyPlayingSongData.id then return end

    obj:setVolumePitch(currentlyPlayingSongData.id, radioVolume * 0.1 * (currentlyPlayingSongData.volCoef or 1), pitch)

    if not setByPause then
        guihooks.trigger('ZeitRadioVolumeUpdate', {volume = radioVolume})
    end

    if hideMessages then return end
    if showMessages then
        if radioVolume == 20 then
            guihooks.message("<span><span style='font-weight: bold;'>Radio Volume: </span>Max</span>", 2, "vehicle.zeitRadio.volumeState", "volume_up")
        else
            guihooks.message("<span><span style='font-weight: bold;'>Radio Volume: </span>"..radioVolume.."</span>", 2, "vehicle.zeitRadio.volumeState", "volume_down")
        end
    end
end

local function pausePlaySong()
    if not isPaused then
        isPaused = true
        prePauseRadioVolume = radioVolume
        pitch = 0
        tuneRadioVolume(0, false, true, true)
    else
        pitch = 1
        isPaused = false
        tuneRadioVolume(prePauseRadioVolume, false, true, true)
    end
    guihooks.trigger('ZeitRadioPauseUpdate', {isPaused = isPaused})
end

local function createSFXSource(url, songId)
    -- hardcoded: we must replace the existing dicta SFXProfile to be able to skip around in it
    if url == "music_menu" then url = "art/sound/music/dicta.ogg" end
    local id = obj:createSFXSource2(url, "AudioRadioMusic2D", tostring(songId), soundNode, -1)

    -- doesnt work
    --obj:setSFXparameter(id, "muffle", "0")
    --obj:setSFXparameter(id, "reverb", "0")
    --obj:setSFXparameter(id, "muffled", "0")
    --obj:setSFXparameter(id, "reverbed", "0")
    obj:stopSFX(id)
    obj:cutSFX(id)
    return id
end

local function unloadSongs(list)
    if not list or not playlists[list] then return end
    for _,v2 in ipairs(playlists[list]) do
        --n=n+1 -- n: amount of songs in table
        if not trackTable[v2] or not trackTable[v2].id or v2 == "_tempData" then goto next end
        obj:stopSFX(trackTable[v2].id or 0)
        obj:cutSFX(trackTable[v2].id or 0)

        obj:deleteSFXSource(trackTable[v2].id)
        obj:setVolumePitch(trackTable[v2].id, 0, 0)
        trackTable[v2].id = nil

        ::next::
    end
end

local function unloadSong(name)
    if not name or not trackTable[name] or not trackTable[name].id then return end
    obj:stopSFX(trackTable[name].id or 0)
    obj:cutSFX(trackTable[name].id or 0)

    obj:deleteSFXSource(trackTable[name].id)
    obj:setVolumePitch(trackTable[name].id, 0, 0)
    trackTable[name].id = nil
end

local function loadSong(name, url)
    local id = createSFXSource(url, name)
    trackTable[name].id = id
    return id
end

local function changeSong(val, fullVal) -- fullVal is a bool!
    if not fullVal then
        currentlyPlayingSongIndex = currentlyPlayingSongIndex + val
    elseif fullVal then
        currentlyPlayingSongIndex = val
    end
    currentlyPlayingSongIndex = math.min(math.max(currentlyPlayingSongIndex, 1), amountOfSongsInTable)
    songTimer = 0

    if currentlyPlayingSongData and currentlyPlayingSongData.id then
        obj:stopSFX(currentlyPlayingSongData.id)
        obj:cutSFX(currentlyPlayingSongData.id)
        table.insert(songUnloadQueue,currentlyPlayingSongData.id)
    end

    currentlyPlayingSongData = trackTable[currentPlaylistIds[currentlyPlayingSongIndex]] or trackTable["_tempData"]

    if not currentlyPlayingSongData then return end
    if not currentlyPlayingSongData.id then
        currentlyPlayingSongData.id = loadSong(currentPlaylistIds[currentlyPlayingSongIndex], currentlyPlayingSongData.url)
    end

    obj:playSFX(currentlyPlayingSongData.id, soundNode, radioVolume, 1)

    tuneRadioVolume(0, false, false, true)

    guihooks.trigger('ZeitRadioSongUpdate', currentlyPlayingSongData)
    if hideMessages then return end
    if currentlyPlayingSongIndex == 1 then
        guihooks.message("<span style='font-weight: bold;'>Radio Off</span>", 2, "vehicle.zeitRadio.song", "volume_off")
    else
        guihooks.message(
            "<span><span style='font-weight: bold;'>Now Playing: </span>"..currentlyPlayingSongData.name.." <span style='font-weight: bold;'> from Playlist: </span> "..currentPlaylistIndex.." </span>",
        2, "vehicle.zeitRadio.song", "volume_up")
    end
end

local function changePlaylist(val, fullVal) -- fullVal is a bool!
    --local lastPlaylistIndex = currentPlaylistIndex
    if not fullVal then
        currentPlaylistIndex = currentPlaylistIndex + val
    elseif fullVal then
        currentPlaylistIndex = val
    end

    --if lastPlaylistIndex ~= currentPlaylistIndex then
    --    --unloadSongs(lastPlaylistIndex)
    --    for _,v in ipairs(playlists[lastPlaylistIndex]) do
    --        if v ~= "_tempData" then
    --            table.insert(songUnloadQueue,v)
    --        end
    --    end
    --end

    currentPlaylistIndex = math.min(math.max(currentPlaylistIndex, 1), amountOfPlaylists)

    currentPlaylistIds = playlists[currentPlaylistIndex]
    amountOfSongsInTable = #currentPlaylistIds

    if val ~= 0 then currentlyPlayingSongIndex = 1 end

    changeSong(0)
    if hideMessages then return end
    guihooks.message("<span><span style='font-weight: bold;'>Selected Playlist: </span> "..currentPlaylistIndex.." </span>", 2, "vehicle.zeitRadio.playlist", "volume_up")
end

local function checkCamInside()
    camPos = obj:getCameraPosition()

    if camPos and vehPos then
        camDistToCar = camPos:distance(vehPos)

        lastIsInCar = isInCar
        isInCar = camDistToCar <= 0.6
        if lastIsInCar ~= isInCar then
            tuneRadioVolume(0, false)
        end

        if not isInCar and currentlyPlayingSongData and currentlyPlayingSongData.id then
            obj:setVolumePitch(currentlyPlayingSongData.id, radioVolume * (clamp(3-camDistToCar, 0, 2)/3)*0.05, pitch)
        end
    end

    if not driverCamId then
        driverCamId = beamstate.nodeNameMap["driver"] or 0
    end
    vehPos = obj:getPosition() + obj:getNodePosition(driverCamId)
end

local function skipTo(seconds)
    if currentlyPlayingSongData.url:match("event:>") then return end

    seconds = seconds*currentlyPlayingSongData.duration
    --[[
    t3d/beamengine creates an SFXProfile obj and then also an SFXSound obj, which is always one ID higher than the SFXProfile
    luckily we can set the SFXProfile's obj name and grab it from GE like that, the """id""" createSFXSource returns is more than useless
    there is no other way to actually get the SFXSound because it is a separate object from the SFXProfile; the only thing it shares is object inheritance
    we cant even match it because the sound it controls isnt exposed

    this doesnt work for event sounds (or really any sound that isnt created by the radio itself), only for externals: hence we just return and ignore it
    the only song that is an event is the lightrunner loop, which in of itself is actually 4 songs and also resets on respawn
    so events get special treatment there as well
    ]]--
    obj:queueGameEngineLua([[
        local sfxProfile = scenetree.findObject(']]..currentPlaylistIds[currentlyPlayingSongIndex]..[[')
        if sfxProfile and sfxProfile.getID then
            local sfxSound = scenetree.findObject(sfxProfile:getID()+1)
            if sfxSound then
                sfxSound:setPosition(]]..seconds..[[)
            end
        end
    ]])
    songTimer = seconds
end

local function updateGFX(dt)
    if songUnloadQueue[#songUnloadQueue] then
        unloadSong(songUnloadQueue[#songUnloadQueue])
        songUnloadQueue[#songUnloadQueue] = nil
    end

    if playerInfo.anyPlayerSeated then
        if not dontCamInCarCheck then
            checkCamInside()
        end

        if not isPaused and currentlyPlayingSongData and currentlyPlayingSongData.duration ~= 0 and type(currentlyPlayingSongData.duration) == "number" then
            songTimer = songTimer + dt
            if songTimer >= currentlyPlayingSongData.duration then
                if settings.getValue("zeit_radio_shuffleplaylist") then
                    changeSong(math.random(2, amountOfSongsInTable), true)
                else
                    local newIndex = (currentlyPlayingSongIndex+1)%amountOfSongsInTable
                    if settings.getValue("zeit_radio_loopplaylist") and newIndex == 1 then
                        newIndex = 2
                    end
                    changeSong(newIndex, true)
                end
            end
        end

        -- Update UI every 15 frames
        updateUITimer = updateUITimer + 1
        if updateUITimer >= 15 and obj:getUpdateUIflag() then
            guihooks.trigger('ZeitRadioTimeUpdate', M.getCurrentlyPlayingSongData())
            updateUITimer = 0
        end
    end
end

-- pause on player change
local isAutoPaused = false
local function onPlayersChanged(playerSeated)
    if not playerSeated and not isPaused and not isAutoPaused then
        pausePlaySong()
        isAutoPaused = true
    elseif playerSeated and isPaused and isAutoPaused then
        pausePlaySong()
        isAutoPaused = false
    end
end

local function toggleSystem(bool)
    if bool == false then
        hideMessages = true
        changePlaylist(1, true)
        changeSong(1, true)
        guihooks.trigger('ZeitRadioAppEnableForceHide') -- enable force hide
    elseif bool == true then
        hideMessages = false
        guihooks.trigger('ZeitRadioAppDisableForceHide') -- disable force hide
    end
end

-- I severly shot myself in the leg making the playlists not an array but rather a dictionary using numbers as keys
-- I believe that is the dumbest thing I've ever done
-- so anyways I'm doing this terribleness to convert it into an array
local function fixArrayDictMistake()
    -- make array with indexes
    local indexes = {}
    for k,_ in pairs(playlists) do
        table.insert(indexes, k)
    end

    -- sort array
    table.sort(indexes)

    local newPlaylists = {}
    local m = 0
    for _,key in ipairs(indexes) do
        m=m+1
        newPlaylists[m] = playlists[key]
    end

    return newPlaylists
end

local function onExtensionLoaded()
    -- very first thing to do is add the custom sfx desc
    obj:queueGameEngineLua([[ loadJsonMaterialsFile("/settings/zeit/radio/AudioRadioMusic2D.json") ]])
    -- use refnode for the sound sources
    soundNode = v.data.refNodes[0].ref

    -- read json
    trackTable = jsonReadFile("/settings/zeit/radio/trackTable.json")
    playlists = jsonReadFile("/settings/zeit/radio/playlists.json") or (trackTable and trackTable["playlists"] or nil)
    if not trackTable or not playlists then
        if FS:fileExists("/settings/zeit/radio/trackTable.json") then
            local file = readFile("/settings/zeit/radio/trackTable.json")
            writeFile("/settings/zeit/radio/trackTable_backup_"..os.date():lower():gsub('[%s]', '_'):gsub(':', '')..".json", file)
        end
        if FS:fileExists("/settings/zeit/radio/playlists.json") then
            local file = readFile("/settings/zeit/radio/playlists.json")
            writeFile("/settings/zeit/radio/playlists_backup_"..os.date():lower():gsub('[%s]', '_'):gsub(':', '')..".json", file)
        end

        trackTable = jsonReadFile("/settings/zeit/radio/trackTableDefault.json")
        playlists = jsonReadFile("/settings/zeit/radio/playlistsDefault.json")
        jsonWriteFile("/settings/zeit/radio/trackTable.json", trackTable, true)
        jsonWriteFile("/settings/zeit/radio/playlists.json", playlists, true)
        log("E", "zeitRadio", "Copying default track table to userfolder as file was not found or invalid")

        -- Do a funky lil top-right corner pop-up message
        guihooks.trigger('toastrMsg', {
            type = "warning",
            title = "Zeit Radio (Zeidio)",
            msg = "Copying the default track table files to userfolder as the files were not found or are invalid. This should only happen once after installing the mod.",
            config = {
                timeOut = 20000,
                progressBar = true,
                closeButton = true,

                -- default stuffs
                positionClass = 'toast-top-right',
                preventDuplicates = true,
                preventOpenDuplicates = true,
            }
        })
    end

    -- tempData is a non-existant song as a way to turn the radio off
    trackTable["_tempData"] = {
        artist = "empty",
        coverUrl = "art/zeitRadio/empty.png",
        duration = 0,
        name = "empty",
        url = "art/zeitRadio/empty.ogg",
        volCoef = 0,
    }

    -- process playlists
    if tableIsDict(playlists) then
        playlists = fixArrayDictMistake()
    end
    playlists[1] = playlists[1] or {}
    for _,v in ipairs(playlists) do
        table.insert(v, 1, "_tempData")
    end

    amountOfPlaylists = #playlists

    changePlaylist(0) -- done to update the radio on reset and first spawn
    --guihooks.message("Radio Off", 2, "vehicle.zeitRadio.state", "volume_off")

    local setting = settings.getValue("zeit_radio_thirdperson")
    dontCamInCarCheck = setting == nil and false or setting

    setting = settings.getValue("zeit_radio_rememberafterreset")
    doSerialize = setting == nil and true or setting

    if controller.mainController.name ~= "vehicleController" then toggleSystem(false) end
end

local function onReset()
    if currentlyPlayingSongData.url:match("event:>") then songTimer = 0 end
    for k,v in pairs(trackTable) do
        if v.id and v.id ~= currentlyPlayingSongData.id then
            obj:stopSFX(v.id)
            obj:cutSFX(v.id)
        end
    end
end

-- save data on vehicle reloads so we can continue where left off
local function onSerialize(type)
    if not doSerialize then return end
    if type == "reload" then
        return {
            vol = radioVolume,
            progress = math.max(songTimer/currentlyPlayingSongData.duration, 0),
            playlist = currentlyPlayingSongData.playlist,
            songIndex = currentlyPlayingSongIndex,
            paused = pitch == 0
        }
    end
end

local function onDeserialize(cacheData)
    if cacheData and cacheData.playlist and cacheData.songIndex and cacheData.vol and cacheData.progress then
        changePlaylist(cacheData.playlist, true)
        changeSong(cacheData.songIndex, true)
        skipTo(cacheData.progress)

        prePauseRadioVolume = cacheData.vol
        isPaused = not cacheData.paused
        pausePlaySong()
    end
end

local function onSettingsChanged()
    local setting = settings.getValue("zeit_radio_thirdperson")
    dontCamInCarCheck = setting == nil and false or setting

    setting = settings.getValue("zeit_radio_rememberafterreset")
    doSerialize = setting == nil and true or setting
    tuneRadioVolume(0, false)
end

-- public interface
M.onReset = onReset
M.onExtensionLoaded = onExtensionLoaded
M.updateGFX = updateGFX

M.onSettingsChanged = onSettingsChanged
M.onSerialize = onSerialize
M.onDeserialize = onDeserialize
M.onPlayersChanged = onPlayersChanged

-- input and ge feedback
M.tuneRadioVolume = tuneRadioVolume
M.changeSong = changeSong
M.changePlaylist = changePlaylist
M.skipTo = skipTo

M.pausePlaySong = pausePlaySong

-- support for outside code (display etc)
function M.getCurrentlyPlayingSongData()
    local data = currentlyPlayingSongData
    data.amountOfSongsInList = amountOfSongsInTable
    data.currentSongIndex = currentlyPlayingSongIndex
    data.timeSinceStart = songTimer
    data.playlist = currentPlaylistIndex
    data.songsInPlaylist = currentPlaylistIds
    data.volume = radioVolume
    data.isPaused = isPaused
    return currentlyPlayingSongData
end
function M.requestState()
    local data = M.getCurrentlyPlayingSongData()
    guihooks.trigger('ZeitRadioTimeUpdate', data)
    guihooks.trigger('ZeitRadioSongUpdate', data)
    guihooks.trigger('ZeitRadioPauseUpdate', {isPaused = data.isPaused})
    if data.isPaused then
        guihooks.trigger('ZeitRadioVolumeUpdate', {volume = prePauseRadioVolume})
    else
        guihooks.trigger('ZeitRadioVolumeUpdate', {volume = data.volume})
    end
    M.getCurrentlyPlayingSongData()
end
function M.getCurrentlyPlayingSongDataSerialized()
    return serialize(M.getCurrentlyPlayingSongData())
end
function M.getSongDataById(id)
    return trackTable[id]
end
function M.getPlaylistSongsById(id)
    return playlists[id]
end

M.toggleSystem = toggleSystem

return M