angular.module('beamng.apps')
.directive('zeitRadioControls', [function () {
  return {
	templateUrl: '/ui/modules/apps/zeitRadioControls/app.html',

  	link: function (scope, element) {
		let allowHideControl = true
		let mainObj = document.getElementById("mainObject")

		let songNameAnim
		let artistNameAnim
		updateTextBoxAnimation = function() {
			if (mainObj.offsetWidth < 400) return
			var textInfoWrap = document.getElementById("nameWrapper").clientWidth + 5

			var songNameObj = document.getElementById("songName")
			if (songNameAnim) songNameAnim.cancel()
			if (textInfoWrap < songNameObj.scrollWidth) {
				var timeModifier = songNameObj.scrollWidth/160
				songNameAnim = songNameObj.animate(
					[
						{transform:`translateX(              ${textInfoWrap}px)`},
						{transform:`translateX(                            0px)`},
						{transform:`translateX(                            0px)`},
						{transform:`translateX(-${songNameObj.scrollWidth+5}px)`}
					], {
						duration: 8000*timeModifier,
						iterations: Infinity,
					},
				)
				songNameAnim.currentTime = 5000*timeModifier
			}

			var artistNameObj = document.getElementById("artistName")
			if (artistNameAnim) artistNameAnim.cancel()
			if (textInfoWrap < artistNameObj.scrollWidth) {
				var timeModifier = artistNameObj.scrollWidth/160
				artistNameAnim = artistNameObj.animate(
					[
						{transform:`translateX(                ${textInfoWrap}px)`},
						{transform:`translateX(                              0px)`},
						{transform:`translateX(                              0px)`},
						{transform:`translateX(-${artistNameObj.scrollWidth+5}px)`}
					], {
						duration: 8000*timeModifier,
						iterations: Infinity,
					},
				)
				artistNameAnim.currentTime = 5000*timeModifier
			}
		}

		scope.$on('app:resized', function (event, data) {
			document.getElementById("nameWrapper").style.display = data.width<440?"none":"inline"
		})

		scope.$on('ZeitRadioTimeUpdate', function (event, data) {
			if (!scope.adjustingTime) scope.songTime = (data.timeSinceStart / data.duration) || 0

			var playedMinutes = Math.trunc(data.timeSinceStart / 60)
			var playedSeconds = (Math.trunc(data.timeSinceStart) - Math.trunc(data.timeSinceStart / 60)*60)

			if (playedSeconds <= 9) { playedSeconds = "0" + playedSeconds}
			document.getElementById("playedTime").innerText = playedMinutes + ":" + playedSeconds

			var totalMinutes = Math.trunc(data.duration / 60)
			var totalSeconds = (Math.trunc(data.duration) - Math.trunc(data.duration / 60)*60)

			if (totalSeconds <= 9) { totalSeconds = "0" + totalSeconds}
			document.getElementById("fullDuration").innerText = totalMinutes + ":" + totalSeconds
		})

		scope.$on('ZeitRadioPauseUpdate', function (event, data) {
			document.getElementById("pauseButton").src = data.isPaused?"/ui/modules/apps/zeitRadioControls/play.png":"/ui/modules/apps/zeitRadioControls/pause.png"
		})

		scope.$on('ZeitRadioVolumeUpdate', function (event, data) {
			scope.volume = data.volume
			scope.volumeStr = data.volume==0?"Off":(data.volume==20?"Full":data.volume)

			let volAdj = document.getElementById("volumeAdjustToggle")
			if (data.volume == 0) {
				volAdj.src = "/ui/modules/apps/zeitRadioControls/volume_off.png"
			} else if (data.volume > 1 && data.volume <= 6) {
				volAdj.src = "/ui/modules/apps/zeitRadioControls/volume_low.png"
			} else if (data.volume > 6 && data.volume <= 13) {
				volAdj.src = "/ui/modules/apps/zeitRadioControls/volume_medium.png"
			} else if (data.volume > 13) {
				volAdj.src = "/ui/modules/apps/zeitRadioControls/volume_full.png"
			}
		})

		scope.$on('ZeitRadioSongUpdate', function (event, data) {
			data.coverUrl = data.coverUrl[0] == "/" ? data.coverUrl : "/"+data.coverUrl
			document.getElementById("cover").src = data.coverUrl
			document.getElementById("blurCover").src = data.coverUrl

			document.getElementById("songName").innerText = data.name=="empty"?"No song selected":data.name
			document.getElementById("artistName").innerText = data.artist=="empty"?"No song selected":data.artist

			updateTextBoxAnimation()
		})

		scope.volumeHidden = true
		scope.toggleVolumeBar = function() {
			document.getElementById("volumeAdjustBackground").style.display = scope.volumeHidden?"flex":"none"
			document.getElementById("timebar").style.filter = scope.volumeHidden?"blur(5px)":""

			scope.volumeHidden = !scope.volumeHidden
		}

		scope.toggleHide = function() {
			if (!allowHideControl) return

			if (scope.hidden) {
				mainObj.style.marginLeft = "0%"
				mainObj.style.opacity = 1

				if (artistNameAnim) artistNameAnim.cancel()
				if (songNameAnim) songNameAnim.cancel()
				setTimeout(function() { updateTextBoxAnimation() }, 1100)
			} else {
				mainObj.style.marginLeft = "100%"
				mainObj.style.opacity = 0
			}
			scope.hidden = !scope.hidden
		}

		scope.$on('ZeitRadioAppEnableForceHide', function (event) {
			if (!scope.hidden) scope.toggleHide()
			allowHideControl = false
		})
		scope.$on('ZeitRadioAppDisableForceHide', function (event) {
			allowHideControl = true
			if (scope.hidden) scope.toggleHide()
		})

		scope.nextList = function() { bngApi.activeObjectLua("if zeitRadio then zeitRadio.changePlaylist(1) end") }
		scope.prevList = function() { bngApi.activeObjectLua("if zeitRadio then zeitRadio.changePlaylist(-1) end") }
		scope.nextSong = function() { bngApi.activeObjectLua("if zeitRadio then zeitRadio.changeSong(1) end") }
		scope.prevSong = function() { bngApi.activeObjectLua("if zeitRadio then zeitRadio.changeSong(-1) end") }
		scope.setVolume = function(volume) { bngApi.activeObjectLua(`if zeitRadio then zeitRadio.tuneRadioVolume(${volume}, true, true) end`) }
		scope.skipTo = function(songTime) { bngApi.activeObjectLua(`if zeitRadio then zeitRadio.skipTo(${songTime}) end`) }
		scope.pausePlaySong = function() { bngApi.activeObjectLua("if zeitRadio then zeitRadio.pausePlaySong() end") }
		scope.toggleEditor = function() { bngApi.engineLua("if zeit_tracktableEditor then extensions.unload('zeit_tracktableEditor') else extensions.load('zeit_tracktableEditor') end") }

		// get data
		bngApi.activeObjectLua('zeitRadio.requestState()')
		// unhide after app load
		scope.hidden = true
		scope.toggleHide()
}}}]);